import chainer
from chainer import link
from chainer import variable
from chainer import functions as F

from chainer_addons.functions import signed_square_root

import numpy as np


class CompactBilinearPooling(link.Link):
	"""
		Compact Bilinear Pooling based on
		https://github.com/ronghanghu/tensorflow_compact_bilinear_pooling
		and https://arxiv.org/abs/1511.06062
	"""
	def __init__(self, input_dim, output_dim, sum_pool=True, normalize=False, seed=2564643, **kw):
		super(CompactBilinearPooling, self).__init__()

		self.input_dim = input_dim
		self.output_dim = output_dim

		rnd = np.random.RandomState(seed)

		h1 = rnd.randint(output_dim, size=input_dim).astype(np.int32)
		s1 = rnd.choice([-1, 1], size=input_dim).astype(np.int32)

		h2 = rnd.randint(output_dim, size=input_dim).astype(np.int32)
		s2 = rnd.choice([-1, 1], size=input_dim).astype(np.int32)

		m1 = self._sketch_matrix(h1, s1)
		m2 = self._sketch_matrix(h2, s2)

		self.add_persistent("sketch_matrix1", m1)
		self.add_persistent("sketch_matrix2", m2)

		self.add_persistent("sum_pool", sum_pool)
		self.add_persistent("normalize", normalize)

	def __call__(self, x1, x2=None):
		if x2 is None:
			x2 = x1

		sketch1 = self._sketch(self.sketch_matrix1, x1)
		sketch2 = self._sketch(self.sketch_matrix2, x2)

		fft_s1, fft_s2 = self._fft(sketch1), self._fft(sketch2)

		res = self._ifft(self._mul_imag(fft_s1, fft_s2))

		if self.sum_pool:
			res = F.mean(res, axis=(1,2))

		if self.normalize:
			res = signed_square_root(res)
			# res = F.sign(res) * F.sqrt(F.absolute(res))
			res = F.normalize(res)

		return res

	def _fft(self, var):
		imag = self.xp.zeros_like(var.data)
		return F.fft((var, imag))

	def _ifft(self, var):
		real, imag = F.ifft(var)
		return real

	def _mul_imag(self, xy, uv):
		x, y = xy
		u, v = uv
		real = x*u - y*v
		imag = x*v + y*u
		return real, imag

	def _sketch_matrix(self, h, s):
		shape = (self.input_dim, self.output_dim)
		res = self.xp.zeros(shape, dtype=np.float32)
		ys = self.xp.arange(self.input_dim)
		res[ys, h] = s
		return res

	def _sketch(self, mat, x):
		x = F.transpose(x, axes=(0,2,3,1))
		x_flat = F.reshape(x, (-1, x.shape[-1]))
		res = F.tensordot(x_flat, mat, axes=1)
		res_shape = x.shape[:-1] + (-1,)
		return F.reshape(res, res_shape)








import chainer
import chainer.links as L

from .base import BaseClassifier

class Classifier(BaseClassifier):
	"""
		model wrapper, that is adapted for the DSL of
		the pretrained models in chainer_addons
	"""

	def __call__(self, X, y):
		pred = self.model(X, layer_name=self.layer_name)
		loss, accu = self.loss(pred, y), self.model.accuracy(pred, y)

		chainer.report({
			"loss": loss.data,
			"accuracy": accu.data,
		}, self)
		return loss

class RNNClassifier(BaseClassifier):
	def __init__(self, model, layer_name=None, rnn_cls=L.LSTM, units=2048, **kwargs):
		layer_name = layer_name or model.meta.feature_layer
		super(RNNClassifier, self).__init__(model, layer_name, **kwargs)
		with self.init_scope():
			self.rnn = rnn_cls(model.meta.feature_size, units)
			self.clf = L.Linear(units, model.clf_layer.W.shape[0])

	def __call__(self, X, y):
		assert X.ndim == 5, "Incorrect input data format!"

		self.rnn.reset_state()

		for part in X.transpose(1,0,2,3,4):
			part_feat = self.model(part, layer_name=self.layer_name)
			feat = self.rnn(part_feat)

		pred = self.clf(feat)
		loss, accu = self.loss(pred, y), self.model.accuracy(pred, y)

		chainer.report({
			"loss": loss.data,
			"accuracy": accu.data,
		}, self)
		return loss



import chainer
import chainer.functions as F

from chainer.serializers import npz
from chainer_addons.links import PoolingType

from collections import OrderedDict
from chainer.initializers import HeNormal

import numpy as np
from abc import ABC, abstractmethod, abstractproperty


class PretrainedModelMixin(ABC):
	CHAINER_PRETRAINED = (
		chainer.links.model.vision.resnet.ResNetLayers,
		chainer.links.model.vision.vgg.VGG16Layers,
		chainer.links.model.vision.googlenet.GoogLeNet,
	)

	def loss(self, pred, gt, loss_func=F.softmax_cross_entropy):
		return loss_func(pred, gt)

	def accuracy(self, pred, gt):
		return F.accuracy(pred, gt)

	def __init__(self, n_classes=1000,
		pooling=PoolingType.Default, pooling_params={}, *args, **kwargs):
		if isinstance(self, PretrainedModelMixin.CHAINER_PRETRAINED):
			super(PretrainedModelMixin, self).__init__(
				pretrained_model=None, *args, **kwargs)
		else:
			super(PretrainedModelMixin, self).__init__(*args, **kwargs)

		if "input_dim" not in pooling_params:
			pooling_params["input_dim"] = self.meta.feature_size

		with self.init_scope():
			self.init_layers(n_classes)
			self.pool = PoolingType.new(pooling, **pooling_params)


	def load_for_finetune(self, weights, n_classes, **kwargs):
		"""
			The weights should be pre-trained on a bigger
			dataset (eg. ImageNet). The classification layer is
			reinitialized after all other weights are loaded
		"""
		self.load(weights)
		self.reinitialize_clf(n_classes, **kwargs)

	def load_for_inference(self, weights, n_classes, **kwargs):
		"""
			In this use case we are loading already fine-tuned
			weights. This means, we need to reinitialize the
			classification layer first and then load the weights.
		"""
		self.reinitialize_clf(n_classes, **kwargs)
		self.load(weights)

	def load(self, weights, strict=False):
		if weights not in [None, "auto"]:
			npz.load_npz(weights, self, strict=strict)


	@property
	def clf_layer(self):
		clf_layer_name = self.meta.classifier_layers[-1]
		return getattr(self, clf_layer_name)

	@classmethod
	def prepare(cls, img, size=None):
		size = size or cls.meta.input_size
		if isinstance(size, int):
			size = (size, size)
		return cls.meta.prepare_func(img, size=size)

	@classmethod
	def prepare_back(cls, img):
		img = img.transpose(1,2,0).copy()
		mean = cls.meta.mean
		if isinstance(mean, np.ndarray):
			mean = mean.squeeze()
		img += mean
		img = img[..., ::-1].astype(np.uint8)
		return img

	def reinitialize_clf(self, n_classes,
		feat_size=None, initializer=None):
		if initializer is None or not callable(initializer):
			initializer = HeNormal(scale=1.0)

		clf_layer = self.clf_layer

		w_shape = (n_classes, feat_size or clf_layer.W.shape[1])
		clf_layer.W.data = np.zeros(w_shape, dtype=np.float32)
		clf_layer.b.data = np.zeros(w_shape[0], dtype=np.float32)
		initializer(clf_layer.W.data)

	@property
	def functions(self):
		return OrderedDict(self._links)

	@abstractmethod
	def init_layers(self, n_classes):
		raise NotImplementedError()

	@abstractproperty
	def _links(self):
		raise NotImplementedError()

	def __call__(self, X, layer_name=None):
		layer_name = layer_name or self.meta.classifier_layers[-1]
		caller = super(PretrainedModelMixin, self).__call__
		activations = caller(X, layers=[layer_name])
		if isinstance(activations, dict):
			activations = activations[layer_name]
		return activations

class BaseClassifier(ABC, chainer.Chain):
	def __init__(self, model, layer_name=None, loss_func=F.softmax_cross_entropy):
		super(BaseClassifier, self).__init__()
		self.layer_name = layer_name or model.meta.classifier_layers[-1]
		self._loss_func = loss_func

		with self.init_scope():
			self.model = model

	def loss(self, pred, y):
		return self.model.loss(pred, y,
			loss_func=self._loss_func)




import numpy as np
from .base import ContiniousLearningRate

class ExponentialLearningRate(ContiniousLearningRate):

	@property
	def factor(self):
		return pow(10, np.log10(self._target / self._lr) / self._epochs)

	def new_lr(self, epoch):
		return self._lr * self.factor ** epoch



import numpy as np
from .base import ContiniousLearningRate

class CosineAnnealingLearningRate(ContiniousLearningRate):

	@property
	def factor(self):
		return 0.5 * (self._lr - self._target)

	def new_lr(self, epoch):
		epochs = self._epochs - self._offset
		return self._target + self.factor * (1 + np.cos(np.pi * (epoch) / epochs))




chainer_to_keras = {
	"/head/conv1/conv": "conv2d_1",
	"/head/conv1/bn": "batch_normalization_1",

	"/head/conv2/conv": "conv2d_2",
	"/head/conv2/bn": "batch_normalization_2",

	"/head/conv3/conv": "conv2d_3",
	"/head/conv3/bn": "batch_normalization_3",

	"/head/conv5/conv": "conv2d_4",
	"/head/conv5/bn": "batch_normalization_4",

	"/head/conv6/conv": "conv2d_5",
	"/head/conv6/bn": "batch_normalization_5",

	"/mixed00/conv1x1/conv": "conv2d_6",
	"/mixed00/conv1x1/bn": "batch_normalization_6",

	"/mixed00/conv5x5_1/conv": "conv2d_7",
	"/mixed00/conv5x5_1/bn": "batch_normalization_7",

	"/mixed00/conv5x5_2/conv": "conv2d_8",
	"/mixed00/conv5x5_2/bn": "batch_normalization_8",

	"/mixed00/conv3x3_1/conv": "conv2d_9",
	"/mixed00/conv3x3_1/bn": "batch_normalization_9",

	"/mixed00/conv3x3_2/conv": "conv2d_10",
	"/mixed00/conv3x3_2/bn": "batch_normalization_10",

	"/mixed00/conv3x3_3/conv": "conv2d_11",
	"/mixed00/conv3x3_3/bn": "batch_normalization_11",

	"/mixed00/pool_conv/conv": "conv2d_12",
	"/mixed00/pool_conv/bn": "batch_normalization_12",


	"/mixed01/conv1x1/conv": "conv2d_13",
	"/mixed01/conv1x1/bn": "batch_normalization_13",

	"/mixed01/conv5x5_1/conv": "conv2d_14",
	"/mixed01/conv5x5_1/bn": "batch_normalization_14",

	"/mixed01/conv5x5_2/conv": "conv2d_15",
	"/mixed01/conv5x5_2/bn": "batch_normalization_15",

	"/mixed01/conv3x3_1/conv": "conv2d_16",
	"/mixed01/conv3x3_1/bn": "batch_normalization_16",

	"/mixed01/conv3x3_2/conv": "conv2d_17",
	"/mixed01/conv3x3_2/bn": "batch_normalization_17",

	"/mixed01/conv3x3_3/conv": "conv2d_18",
	"/mixed01/conv3x3_3/bn": "batch_normalization_18",

	"/mixed01/pool_conv/conv": "conv2d_19",
	"/mixed01/pool_conv/bn": "batch_normalization_19",


	"/mixed02/conv1x1/conv": "conv2d_20",
	"/mixed02/conv1x1/bn": "batch_normalization_20",

	"/mixed02/conv5x5_1/conv": "conv2d_21",
	"/mixed02/conv5x5_1/bn": "batch_normalization_21",

	"/mixed02/conv5x5_2/conv": "conv2d_22",
	"/mixed02/conv5x5_2/bn": "batch_normalization_22",

	"/mixed02/conv3x3_1/conv": "conv2d_23",
	"/mixed02/conv3x3_1/bn": "batch_normalization_23",

	"/mixed02/conv3x3_2/conv": "conv2d_24",
	"/mixed02/conv3x3_2/bn": "batch_normalization_24",

	"/mixed02/conv3x3_3/conv": "conv2d_25",
	"/mixed02/conv3x3_3/bn": "batch_normalization_25",

	"/mixed02/pool_conv/conv": "conv2d_26",
	"/mixed02/pool_conv/bn": "batch_normalization_26",


	"/mixed03/conv3x3/conv": "conv2d_27",
	"/mixed03/conv3x3/bn": "batch_normalization_27",

	"/mixed03/conv3x3_1/conv": "conv2d_28",
	"/mixed03/conv3x3_1/bn": "batch_normalization_28",

	"/mixed03/conv3x3_2/conv": "conv2d_29",
	"/mixed03/conv3x3_2/bn": "batch_normalization_29",

	"/mixed03/conv3x3_3/conv": "conv2d_30",
	"/mixed03/conv3x3_3/bn": "batch_normalization_30",


	"/mixed04/conv1x1/conv": "conv2d_31",
	"/mixed04/conv1x1/bn": "batch_normalization_31",

	"/mixed04/conv7x7_1/conv": "conv2d_32",
	"/mixed04/conv7x7_1/bn": "batch_normalization_32",

	"/mixed04/conv7x7_2/conv": "conv2d_33",
	"/mixed04/conv7x7_2/bn": "batch_normalization_33",

	"/mixed04/conv7x7_3/conv": "conv2d_34",
	"/mixed04/conv7x7_3/bn": "batch_normalization_34",

	"/mixed04/conv7x7x2_1/conv": "conv2d_35",
	"/mixed04/conv7x7x2_1/bn": "batch_normalization_35",

	"/mixed04/conv7x7x2_2/conv": "conv2d_36",
	"/mixed04/conv7x7x2_2/bn": "batch_normalization_36",

	"/mixed04/conv7x7x2_3/conv": "conv2d_37",
	"/mixed04/conv7x7x2_3/bn": "batch_normalization_37",

	"/mixed04/conv7x7x2_4/conv": "conv2d_38",
	"/mixed04/conv7x7x2_4/bn": "batch_normalization_38",

	"/mixed04/conv7x7x2_5/conv": "conv2d_39",
	"/mixed04/conv7x7x2_5/bn": "batch_normalization_39",

	"/mixed04/pool_conv/conv": "conv2d_40",
	"/mixed04/pool_conv/bn": "batch_normalization_40",


	"/mixed05/conv1x1/conv": "conv2d_41",
	"/mixed05/conv1x1/bn": "batch_normalization_41",

	"/mixed05/conv7x7_1/conv": "conv2d_42",
	"/mixed05/conv7x7_1/bn": "batch_normalization_42",

	"/mixed05/conv7x7_2/conv": "conv2d_43",
	"/mixed05/conv7x7_2/bn": "batch_normalization_43",

	"/mixed05/conv7x7_3/conv": "conv2d_44",
	"/mixed05/conv7x7_3/bn": "batch_normalization_44",

	"/mixed05/conv7x7x2_1/conv": "conv2d_45",
	"/mixed05/conv7x7x2_1/bn": "batch_normalization_45",

	"/mixed05/conv7x7x2_2/conv": "conv2d_46",
	"/mixed05/conv7x7x2_2/bn": "batch_normalization_46",

	"/mixed05/conv7x7x2_3/conv": "conv2d_47",
	"/mixed05/conv7x7x2_3/bn": "batch_normalization_47",

	"/mixed05/conv7x7x2_4/conv": "conv2d_48",
	"/mixed05/conv7x7x2_4/bn": "batch_normalization_48",

	"/mixed05/conv7x7x2_5/conv": "conv2d_49",
	"/mixed05/conv7x7x2_5/bn": "batch_normalization_49",

	"/mixed05/pool_conv/conv": "conv2d_50",
	"/mixed05/pool_conv/bn": "batch_normalization_50",


	"/mixed06/conv1x1/conv": "conv2d_51",
	"/mixed06/conv1x1/bn": "batch_normalization_51",

	"/mixed06/conv7x7_1/conv": "conv2d_52",
	"/mixed06/conv7x7_1/bn": "batch_normalization_52",

	"/mixed06/conv7x7_2/conv": "conv2d_53",
	"/mixed06/conv7x7_2/bn": "batch_normalization_53",

	"/mixed06/conv7x7_3/conv": "conv2d_54",
	"/mixed06/conv7x7_3/bn": "batch_normalization_54",

	"/mixed06/conv7x7x2_1/conv": "conv2d_55",
	"/mixed06/conv7x7x2_1/bn": "batch_normalization_55",

	"/mixed06/conv7x7x2_2/conv": "conv2d_56",
	"/mixed06/conv7x7x2_2/bn": "batch_normalization_56",

	"/mixed06/conv7x7x2_3/conv": "conv2d_57",
	"/mixed06/conv7x7x2_3/bn": "batch_normalization_57",

	"/mixed06/conv7x7x2_4/conv": "conv2d_58",
	"/mixed06/conv7x7x2_4/bn": "batch_normalization_58",

	"/mixed06/conv7x7x2_5/conv": "conv2d_59",
	"/mixed06/conv7x7x2_5/bn": "batch_normalization_59",

	"/mixed06/pool_conv/conv": "conv2d_60",
	"/mixed06/pool_conv/bn": "batch_normalization_60",


	"/mixed07/conv1x1/conv": "conv2d_61",
	"/mixed07/conv1x1/bn": "batch_normalization_61",

	"/mixed07/conv7x7_1/conv": "conv2d_62",
	"/mixed07/conv7x7_1/bn": "batch_normalization_62",

	"/mixed07/conv7x7_2/conv": "conv2d_63",
	"/mixed07/conv7x7_2/bn": "batch_normalization_63",

	"/mixed07/conv7x7_3/conv": "conv2d_64",
	"/mixed07/conv7x7_3/bn": "batch_normalization_64",

	"/mixed07/conv7x7x2_1/conv": "conv2d_65",
	"/mixed07/conv7x7x2_1/bn": "batch_normalization_65",

	"/mixed07/conv7x7x2_2/conv": "conv2d_66",
	"/mixed07/conv7x7x2_2/bn": "batch_normalization_66",

	"/mixed07/conv7x7x2_3/conv": "conv2d_67",
	"/mixed07/conv7x7x2_3/bn": "batch_normalization_67",

	"/mixed07/conv7x7x2_4/conv": "conv2d_68",
	"/mixed07/conv7x7x2_4/bn": "batch_normalization_68",

	"/mixed07/conv7x7x2_5/conv": "conv2d_69",
	"/mixed07/conv7x7x2_5/bn": "batch_normalization_69",

	"/mixed07/pool_conv/conv": "conv2d_70",
	"/mixed07/pool_conv/bn": "batch_normalization_70",


	"/mixed08/conv3x3_1/conv": "conv2d_71",
	"/mixed08/conv3x3_1/bn": "batch_normalization_71",

	"/mixed08/conv3x3_2/conv": "conv2d_72",
	"/mixed08/conv3x3_2/bn": "batch_normalization_72",

	"/mixed08/conv7x7_1/conv": "conv2d_73",
	"/mixed08/conv7x7_1/bn": "batch_normalization_73",

	"/mixed08/conv7x7_2/conv": "conv2d_74",
	"/mixed08/conv7x7_2/bn": "batch_normalization_74",

	"/mixed08/conv7x7_3/conv": "conv2d_75",
	"/mixed08/conv7x7_3/bn": "batch_normalization_75",

	"/mixed08/conv7x7_4/conv": "conv2d_76",
	"/mixed08/conv7x7_4/bn": "batch_normalization_76",


	"/mixed09/conv1x1/conv": "conv2d_77",
	"/mixed09/conv1x1/bn": "batch_normalization_77",

	"/mixed09/conv3x3_1/conv": "conv2d_78",
	"/mixed09/conv3x3_1/bn": "batch_normalization_78",

	"/mixed09/conv3x3_2/conv": "conv2d_79",
	"/mixed09/conv3x3_2/bn": "batch_normalization_79",

	"/mixed09/conv3x3_3/conv": "conv2d_80",
	"/mixed09/conv3x3_3/bn": "batch_normalization_80",

	"/mixed09/conv3x3x2_1/conv": "conv2d_81",
	"/mixed09/conv3x3x2_1/bn": "batch_normalization_81",

	"/mixed09/conv3x3x2_2/conv": "conv2d_82",
	"/mixed09/conv3x3x2_2/bn": "batch_normalization_82",

	"/mixed09/conv3x3x2_3/conv": "conv2d_83",
	"/mixed09/conv3x3x2_3/bn": "batch_normalization_83",

	"/mixed09/conv3x3x2_4/conv": "conv2d_84",
	"/mixed09/conv3x3x2_4/bn": "batch_normalization_84",

	"/mixed09/pool_conv/conv": "conv2d_85",
	"/mixed09/pool_conv/bn": "batch_normalization_85",


	"/mixed10/conv1x1/conv": "conv2d_86",
	"/mixed10/conv1x1/bn": "batch_normalization_86",

	"/mixed10/conv3x3_1/conv": "conv2d_87",
	"/mixed10/conv3x3_1/bn": "batch_normalization_87",

	"/mixed10/conv3x3_2/conv": "conv2d_88",
	"/mixed10/conv3x3_2/bn": "batch_normalization_88",

	"/mixed10/conv3x3_3/conv": "conv2d_89",
	"/mixed10/conv3x3_3/bn": "batch_normalization_89",

	"/mixed10/conv3x3x2_1/conv": "conv2d_90",
	"/mixed10/conv3x3x2_1/bn": "batch_normalization_90",

	"/mixed10/conv3x3x2_2/conv": "conv2d_91",
	"/mixed10/conv3x3x2_2/bn": "batch_normalization_91",

	"/mixed10/conv3x3x2_3/conv": "conv2d_92",
	"/mixed10/conv3x3x2_3/bn": "batch_normalization_92",

	"/mixed10/conv3x3x2_4/conv": "conv2d_93",
	"/mixed10/conv3x3x2_4/bn": "batch_normalization_93",

	"/mixed10/pool_conv/conv": "conv2d_94",
	"/mixed10/pool_conv/bn": "batch_normalization_94",

	# "/fc": "predictions",
	"/fc": "prob",

}

chainer_to_tf_ckpt = {
	"/head/conv1/conv": "Conv2d_1a_3x3",
	"/head/conv1/bn": "Conv2d_1a_3x3/BatchNorm",

	"/head/conv2/conv": "Conv2d_2a_3x3",
	"/head/conv2/bn": "Conv2d_2a_3x3/BatchNorm",

	"/head/conv3/conv": "Conv2d_2b_3x3",
	"/head/conv3/bn": "Conv2d_2b_3x3/BatchNorm",

	"/head/conv5/conv": "Conv2d_3b_1x1",
	"/head/conv5/bn": "Conv2d_3b_1x1/BatchNorm",

	"/head/conv6/conv": "Conv2d_4a_3x3",
	"/head/conv6/bn": "Conv2d_4a_3x3/BatchNorm",

	"/mixed00/conv1x1/conv": "Mixed_5b/Branch_0/Conv2d_0a_1x1",
	"/mixed00/conv1x1/bn": "Mixed_5b/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed00/conv5x5_1/conv": "Mixed_5b/Branch_1/Conv2d_0a_1x1",
	"/mixed00/conv5x5_1/bn": "Mixed_5b/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed00/conv5x5_2/conv": "Mixed_5b/Branch_1/Conv2d_0b_5x5",
	"/mixed00/conv5x5_2/bn": "Mixed_5b/Branch_1/Conv2d_0b_5x5/BatchNorm",

	"/mixed00/conv3x3_1/conv": "Mixed_5b/Branch_2/Conv2d_0a_1x1",
	"/mixed00/conv3x3_1/bn": "Mixed_5b/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed00/conv3x3_2/conv": "Mixed_5b/Branch_2/Conv2d_0b_3x3",
	"/mixed00/conv3x3_2/bn": "Mixed_5b/Branch_2/Conv2d_0b_3x3/BatchNorm",

	"/mixed00/conv3x3_3/conv": "Mixed_5b/Branch_2/Conv2d_0c_3x3",
	"/mixed00/conv3x3_3/bn": "Mixed_5b/Branch_2/Conv2d_0c_3x3/BatchNorm",

	"/mixed00/pool_conv/conv": "Mixed_5b/Branch_3/Conv2d_0b_1x1",
	"/mixed00/pool_conv/bn": "Mixed_5b/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed01/conv1x1/conv": "Mixed_5c/Branch_0/Conv2d_0a_1x1",
	"/mixed01/conv1x1/bn": "Mixed_5c/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed01/conv5x5_1/conv": "Mixed_5c/Branch_1/Conv2d_0b_1x1",
	"/mixed01/conv5x5_1/bn": "Mixed_5c/Branch_1/Conv2d_0b_1x1/BatchNorm",

	"/mixed01/conv5x5_2/conv": "Mixed_5c/Branch_1/Conv_1_0c_5x5",
	"/mixed01/conv5x5_2/bn": "Mixed_5c/Branch_1/Conv_1_0c_5x5/BatchNorm",

	"/mixed01/conv3x3_1/conv": "Mixed_5c/Branch_2/Conv2d_0a_1x1",
	"/mixed01/conv3x3_1/bn": "Mixed_5c/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed01/conv3x3_2/conv": "Mixed_5c/Branch_2/Conv2d_0b_3x3",
	"/mixed01/conv3x3_2/bn": "Mixed_5c/Branch_2/Conv2d_0b_3x3/BatchNorm",

	"/mixed01/conv3x3_3/conv": "Mixed_5c/Branch_2/Conv2d_0c_3x3",
	"/mixed01/conv3x3_3/bn": "Mixed_5c/Branch_2/Conv2d_0c_3x3/BatchNorm",

	"/mixed01/pool_conv/conv": "Mixed_5c/Branch_3/Conv2d_0b_1x1",
	"/mixed01/pool_conv/bn": "Mixed_5c/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed02/conv1x1/conv": "Mixed_5d/Branch_0/Conv2d_0a_1x1",
	"/mixed02/conv1x1/bn": "Mixed_5d/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed02/conv5x5_1/conv": "Mixed_5d/Branch_1/Conv2d_0a_1x1",
	"/mixed02/conv5x5_1/bn": "Mixed_5d/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed02/conv5x5_2/conv": "Mixed_5d/Branch_1/Conv2d_0b_5x5",
	"/mixed02/conv5x5_2/bn": "Mixed_5d/Branch_1/Conv2d_0b_5x5/BatchNorm",

	"/mixed02/conv3x3_1/conv": "Mixed_5d/Branch_2/Conv2d_0a_1x1",
	"/mixed02/conv3x3_1/bn": "Mixed_5d/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed02/conv3x3_2/conv": "Mixed_5d/Branch_2/Conv2d_0b_3x3",
	"/mixed02/conv3x3_2/bn": "Mixed_5d/Branch_2/Conv2d_0b_3x3/BatchNorm",

	"/mixed02/conv3x3_3/conv": "Mixed_5d/Branch_2/Conv2d_0c_3x3",
	"/mixed02/conv3x3_3/bn": "Mixed_5d/Branch_2/Conv2d_0c_3x3/BatchNorm",

	"/mixed02/pool_conv/conv": "Mixed_5d/Branch_3/Conv2d_0b_1x1",
	"/mixed02/pool_conv/bn": "Mixed_5d/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed03/conv3x3/conv": "Mixed_6a/Branch_0/Conv2d_1a_1x1",
	"/mixed03/conv3x3/bn": "Mixed_6a/Branch_0/Conv2d_1a_1x1/BatchNorm",

	"/mixed03/conv3x3_1/conv": "Mixed_6a/Branch_1/Conv2d_0a_1x1",
	"/mixed03/conv3x3_1/bn": "Mixed_6a/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed03/conv3x3_2/conv": "Mixed_6a/Branch_1/Conv2d_0b_3x3",
	"/mixed03/conv3x3_2/bn": "Mixed_6a/Branch_1/Conv2d_0b_3x3/BatchNorm",

	"/mixed03/conv3x3_3/conv": "Mixed_6a/Branch_1/Conv2d_1a_1x1",
	"/mixed03/conv3x3_3/bn": "Mixed_6a/Branch_1/Conv2d_1a_1x1/BatchNorm",


	"/mixed04/conv1x1/conv": "Mixed_6b/Branch_0/Conv2d_0a_1x1",
	"/mixed04/conv1x1/bn": "Mixed_6b/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed04/conv7x7_1/conv": "Mixed_6b/Branch_1/Conv2d_0a_1x1",
	"/mixed04/conv7x7_1/bn": "Mixed_6b/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed04/conv7x7_2/conv": "Mixed_6b/Branch_1/Conv2d_0b_1x7",
	"/mixed04/conv7x7_2/bn": "Mixed_6b/Branch_1/Conv2d_0b_1x7/BatchNorm",

	"/mixed04/conv7x7_3/conv": "Mixed_6b/Branch_1/Conv2d_0c_7x1",
	"/mixed04/conv7x7_3/bn": "Mixed_6b/Branch_1/Conv2d_0c_7x1/BatchNorm",

	"/mixed04/conv7x7x2_1/conv": "Mixed_6b/Branch_2/Conv2d_0a_1x1",
	"/mixed04/conv7x7x2_1/bn": "Mixed_6b/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed04/conv7x7x2_2/conv": "Mixed_6b/Branch_2/Conv2d_0b_7x1",
	"/mixed04/conv7x7x2_2/bn": "Mixed_6b/Branch_2/Conv2d_0b_7x1/BatchNorm",

	"/mixed04/conv7x7x2_3/conv": "Mixed_6b/Branch_2/Conv2d_0c_1x7",
	"/mixed04/conv7x7x2_3/bn": "Mixed_6b/Branch_2/Conv2d_0c_1x7/BatchNorm",

	"/mixed04/conv7x7x2_4/conv": "Mixed_6b/Branch_2/Conv2d_0d_7x1",
	"/mixed04/conv7x7x2_4/bn": "Mixed_6b/Branch_2/Conv2d_0d_7x1/BatchNorm",

	"/mixed04/conv7x7x2_5/conv": "Mixed_6b/Branch_2/Conv2d_0e_1x7",
	"/mixed04/conv7x7x2_5/bn": "Mixed_6b/Branch_2/Conv2d_0e_1x7/BatchNorm",

	"/mixed04/pool_conv/conv": "Mixed_6b/Branch_3/Conv2d_0b_1x1",
	"/mixed04/pool_conv/bn": "Mixed_6b/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed05/conv1x1/conv": "Mixed_6c/Branch_0/Conv2d_0a_1x1",
	"/mixed05/conv1x1/bn": "Mixed_6c/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed05/conv7x7_1/conv": "Mixed_6c/Branch_1/Conv2d_0a_1x1",
	"/mixed05/conv7x7_1/bn": "Mixed_6c/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed05/conv7x7_2/conv": "Mixed_6c/Branch_1/Conv2d_0b_1x7",
	"/mixed05/conv7x7_2/bn": "Mixed_6c/Branch_1/Conv2d_0b_1x7/BatchNorm",

	"/mixed05/conv7x7_3/conv": "Mixed_6c/Branch_1/Conv2d_0c_7x1",
	"/mixed05/conv7x7_3/bn": "Mixed_6c/Branch_1/Conv2d_0c_7x1/BatchNorm",

	"/mixed05/conv7x7x2_1/conv": "Mixed_6c/Branch_2/Conv2d_0a_1x1",
	"/mixed05/conv7x7x2_1/bn": "Mixed_6c/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed05/conv7x7x2_2/conv": "Mixed_6c/Branch_2/Conv2d_0b_7x1",
	"/mixed05/conv7x7x2_2/bn": "Mixed_6c/Branch_2/Conv2d_0b_7x1/BatchNorm",

	"/mixed05/conv7x7x2_3/conv": "Mixed_6c/Branch_2/Conv2d_0c_1x7",
	"/mixed05/conv7x7x2_3/bn": "Mixed_6c/Branch_2/Conv2d_0c_1x7/BatchNorm",

	"/mixed05/conv7x7x2_4/conv": "Mixed_6c/Branch_2/Conv2d_0d_7x1",
	"/mixed05/conv7x7x2_4/bn": "Mixed_6c/Branch_2/Conv2d_0d_7x1/BatchNorm",

	"/mixed05/conv7x7x2_5/conv": "Mixed_6c/Branch_2/Conv2d_0e_1x7",
	"/mixed05/conv7x7x2_5/bn": "Mixed_6c/Branch_2/Conv2d_0e_1x7/BatchNorm",

	"/mixed05/pool_conv/conv": "Mixed_6c/Branch_3/Conv2d_0b_1x1",
	"/mixed05/pool_conv/bn": "Mixed_6c/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed06/conv1x1/conv": "Mixed_6d/Branch_0/Conv2d_0a_1x1",
	"/mixed06/conv1x1/bn": "Mixed_6d/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed06/conv7x7_1/conv": "Mixed_6d/Branch_1/Conv2d_0a_1x1",
	"/mixed06/conv7x7_1/bn": "Mixed_6d/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed06/conv7x7_2/conv": "Mixed_6d/Branch_1/Conv2d_0b_1x7",
	"/mixed06/conv7x7_2/bn": "Mixed_6d/Branch_1/Conv2d_0b_1x7/BatchNorm",

	"/mixed06/conv7x7_3/conv": "Mixed_6d/Branch_1/Conv2d_0c_7x1",
	"/mixed06/conv7x7_3/bn": "Mixed_6d/Branch_1/Conv2d_0c_7x1/BatchNorm",

	"/mixed06/conv7x7x2_1/conv": "Mixed_6d/Branch_2/Conv2d_0a_1x1",
	"/mixed06/conv7x7x2_1/bn": "Mixed_6d/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed06/conv7x7x2_2/conv": "Mixed_6d/Branch_2/Conv2d_0b_7x1",
	"/mixed06/conv7x7x2_2/bn": "Mixed_6d/Branch_2/Conv2d_0b_7x1/BatchNorm",

	"/mixed06/conv7x7x2_3/conv": "Mixed_6d/Branch_2/Conv2d_0c_1x7",
	"/mixed06/conv7x7x2_3/bn": "Mixed_6d/Branch_2/Conv2d_0c_1x7/BatchNorm",

	"/mixed06/conv7x7x2_4/conv": "Mixed_6d/Branch_2/Conv2d_0d_7x1",
	"/mixed06/conv7x7x2_4/bn": "Mixed_6d/Branch_2/Conv2d_0d_7x1/BatchNorm",

	"/mixed06/conv7x7x2_5/conv": "Mixed_6d/Branch_2/Conv2d_0e_1x7",
	"/mixed06/conv7x7x2_5/bn": "Mixed_6d/Branch_2/Conv2d_0e_1x7/BatchNorm",

	"/mixed06/pool_conv/conv": "Mixed_6d/Branch_3/Conv2d_0b_1x1",
	"/mixed06/pool_conv/bn": "Mixed_6d/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed07/conv1x1/conv": "Mixed_6e/Branch_0/Conv2d_0a_1x1",
	"/mixed07/conv1x1/bn": "Mixed_6e/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed07/conv7x7_1/conv": "Mixed_6e/Branch_1/Conv2d_0a_1x1",
	"/mixed07/conv7x7_1/bn": "Mixed_6e/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed07/conv7x7_2/conv": "Mixed_6e/Branch_1/Conv2d_0b_1x7",
	"/mixed07/conv7x7_2/bn": "Mixed_6e/Branch_1/Conv2d_0b_1x7/BatchNorm",

	"/mixed07/conv7x7_3/conv": "Mixed_6e/Branch_1/Conv2d_0c_7x1",
	"/mixed07/conv7x7_3/bn": "Mixed_6e/Branch_1/Conv2d_0c_7x1/BatchNorm",

	"/mixed07/conv7x7x2_1/conv": "Mixed_6e/Branch_2/Conv2d_0a_1x1",
	"/mixed07/conv7x7x2_1/bn": "Mixed_6e/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed07/conv7x7x2_2/conv": "Mixed_6e/Branch_2/Conv2d_0b_7x1",
	"/mixed07/conv7x7x2_2/bn": "Mixed_6e/Branch_2/Conv2d_0b_7x1/BatchNorm",

	"/mixed07/conv7x7x2_3/conv": "Mixed_6e/Branch_2/Conv2d_0c_1x7",
	"/mixed07/conv7x7x2_3/bn": "Mixed_6e/Branch_2/Conv2d_0c_1x7/BatchNorm",

	"/mixed07/conv7x7x2_4/conv": "Mixed_6e/Branch_2/Conv2d_0d_7x1",
	"/mixed07/conv7x7x2_4/bn": "Mixed_6e/Branch_2/Conv2d_0d_7x1/BatchNorm",

	"/mixed07/conv7x7x2_5/conv": "Mixed_6e/Branch_2/Conv2d_0e_1x7",
	"/mixed07/conv7x7x2_5/bn": "Mixed_6e/Branch_2/Conv2d_0e_1x7/BatchNorm",

	"/mixed07/pool_conv/conv": "Mixed_6e/Branch_3/Conv2d_0b_1x1",
	"/mixed07/pool_conv/bn": "Mixed_6e/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed08/conv3x3_1/conv": "Mixed_7a/Branch_0/Conv2d_0a_1x1",
	"/mixed08/conv3x3_1/bn": "Mixed_7a/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed08/conv3x3_2/conv": "Mixed_7a/Branch_0/Conv2d_1a_3x3",
	"/mixed08/conv3x3_2/bn": "Mixed_7a/Branch_0/Conv2d_1a_3x3/BatchNorm",

	"/mixed08/conv7x7_1/conv": "Mixed_7a/Branch_1/Conv2d_0a_1x1",
	"/mixed08/conv7x7_1/bn": "Mixed_7a/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed08/conv7x7_2/conv": "Mixed_7a/Branch_1/Conv2d_0b_1x7",
	"/mixed08/conv7x7_2/bn": "Mixed_7a/Branch_1/Conv2d_0b_1x7/BatchNorm",

	"/mixed08/conv7x7_3/conv": "Mixed_7a/Branch_1/Conv2d_0c_7x1",
	"/mixed08/conv7x7_3/bn": "Mixed_7a/Branch_1/Conv2d_0c_7x1/BatchNorm",

	"/mixed08/conv7x7_4/conv": "Mixed_7a/Branch_1/Conv2d_1a_3x3",
	"/mixed08/conv7x7_4/bn": "Mixed_7a/Branch_1/Conv2d_1a_3x3/BatchNorm",


	"/mixed09/conv1x1/conv": "Mixed_7b/Branch_0/Conv2d_0a_1x1",
	"/mixed09/conv1x1/bn": "Mixed_7b/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed09/conv3x3_1/conv": "Mixed_7b/Branch_1/Conv2d_0a_1x1",
	"/mixed09/conv3x3_1/bn": "Mixed_7b/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed09/conv3x3_2/conv": "Mixed_7b/Branch_1/Conv2d_0b_1x3",
	"/mixed09/conv3x3_2/bn": "Mixed_7b/Branch_1/Conv2d_0b_1x3/BatchNorm",

	"/mixed09/conv3x3_3/conv": "Mixed_7b/Branch_1/Conv2d_0b_3x1",
	"/mixed09/conv3x3_3/bn": "Mixed_7b/Branch_1/Conv2d_0b_3x1/BatchNorm",

	"/mixed09/conv3x3x2_1/conv": "Mixed_7b/Branch_2/Conv2d_0a_1x1",
	"/mixed09/conv3x3x2_1/bn": "Mixed_7b/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed09/conv3x3x2_2/conv": "Mixed_7b/Branch_2/Conv2d_0b_3x3",
	"/mixed09/conv3x3x2_2/bn": "Mixed_7b/Branch_2/Conv2d_0b_3x3/BatchNorm",

	"/mixed09/conv3x3x2_3/conv": "Mixed_7b/Branch_2/Conv2d_0c_1x3",
	"/mixed09/conv3x3x2_3/bn": "Mixed_7b/Branch_2/Conv2d_0c_1x3/BatchNorm",

	"/mixed09/conv3x3x2_4/conv": "Mixed_7b/Branch_2/Conv2d_0d_3x1",
	"/mixed09/conv3x3x2_4/bn": "Mixed_7b/Branch_2/Conv2d_0d_3x1/BatchNorm",

	"/mixed09/pool_conv/conv": "Mixed_7b/Branch_3/Conv2d_0b_1x1",
	"/mixed09/pool_conv/bn": "Mixed_7b/Branch_3/Conv2d_0b_1x1/BatchNorm",


	"/mixed10/conv1x1/conv": "Mixed_7c/Branch_0/Conv2d_0a_1x1",
	"/mixed10/conv1x1/bn": "Mixed_7c/Branch_0/Conv2d_0a_1x1/BatchNorm",

	"/mixed10/conv3x3_1/conv": "Mixed_7c/Branch_1/Conv2d_0a_1x1",
	"/mixed10/conv3x3_1/bn": "Mixed_7c/Branch_1/Conv2d_0a_1x1/BatchNorm",

	"/mixed10/conv3x3_2/conv": "Mixed_7c/Branch_1/Conv2d_0b_1x3",
	"/mixed10/conv3x3_2/bn": "Mixed_7c/Branch_1/Conv2d_0b_1x3/BatchNorm",

	"/mixed10/conv3x3_3/conv": "Mixed_7c/Branch_1/Conv2d_0c_3x1",
	"/mixed10/conv3x3_3/bn": "Mixed_7c/Branch_1/Conv2d_0c_3x1/BatchNorm",

	"/mixed10/conv3x3x2_1/conv": "Mixed_7c/Branch_2/Conv2d_0a_1x1",
	"/mixed10/conv3x3x2_1/bn": "Mixed_7c/Branch_2/Conv2d_0a_1x1/BatchNorm",

	"/mixed10/conv3x3x2_2/conv": "Mixed_7c/Branch_2/Conv2d_0b_3x3",
	"/mixed10/conv3x3x2_2/bn": "Mixed_7c/Branch_2/Conv2d_0b_3x3/BatchNorm",

	"/mixed10/conv3x3x2_3/conv": "Mixed_7c/Branch_2/Conv2d_0c_1x3",
	"/mixed10/conv3x3x2_3/bn": "Mixed_7c/Branch_2/Conv2d_0c_1x3/BatchNorm",

	"/mixed10/conv3x3x2_4/conv": "Mixed_7c/Branch_2/Conv2d_0d_3x1",
	"/mixed10/conv3x3x2_4/bn": "Mixed_7c/Branch_2/Conv2d_0d_3x1/BatchNorm",

	"/mixed10/pool_conv/conv": "Mixed_7c/Branch_3/Conv2d_0b_1x1",
	"/mixed10/pool_conv/bn": "Mixed_7c/Branch_3/Conv2d_0b_1x1/BatchNorm",

	# "/fc": "Logits/Conv2d_1c_1x1",

}



import chainer
from chainer import link
from chainer import function
from chainer import initializers
from chainer import functions as F
from chainer.variable import Parameter

import numpy as np

class FVLayer(link.Link):
	"""Fisher Vector Encoding Layer"""
	def __init__(self,
		in_size, n_components,
		init_sigmas=1,
		eps=1e-6):
		super(FVLayer, self).__init__()

		self.n_components = n_components
		self.in_size = in_size

		# init mu with uniform in a range, e.g. [-1, 1] ?
		mu_init = initializers.Uniform(scale=1, dtype=np.float32)
		w_init = initializers.Constant(1 / n_components, dtype=np.float32)
		sig_init = initializers.Constant(init_sigmas, dtype=np.float32)

		with self.init_scope():
			self.mu = Parameter(mu_init,
				shape=(in_size, n_components),
				name="mu")

			self.sig = Parameter(sig_init,
				shape=(in_size, n_components),
				name="sigma")

			self.w = Parameter(w_init,
				shape=(n_components,),
				name="weights")

			self.add_persistent("eps", eps)

		self.two_sq_pi = np.sqrt(2 * np.pi)

	def _expand_params(self, shape):
		_f = lambda param: \
			F.broadcast_to(
				F.expand_dims(
					F.expand_dims(param, axis=0), axis=0), shape)

		return map(_f, [self.mu, self.sig, F.expand_dims(self.w, axis=0)])

	def __call__(self, x):
		n, t, _in_size = x.shape

		shape = (n, t, self.in_size, self.n_components)
		shape2 = (n, self.in_size, self.n_components)

		_x = F.broadcast_to(F.expand_dims(x, -1), shape)
		_mu, _sig, _w = self._expand_params(shape)
		_w2 = F.broadcast_to(
				F.expand_dims(
					F.expand_dims(self.w, axis=0), axis=0), shape2)

		# for numerical stability
		_sig = F.absolute(_sig) + self.eps
		_w = F.elu(_w) + 2 + self.eps
		_w2 = F.elu(_w2) + 2 + self.eps

		# now let the magic begin!

		_g = (_x - _mu) / _sig
		_u = F.exp(-0.5 * _g**2) / (self.two_sq_pi * _sig)

		_wu =  _u * _w
		_wu_sum = F.broadcast_to(F.sum(_wu, axis=2, keepdims=True), shape)

		_gamma = _wu / (_wu_sum + self.eps)


		G_mu = F.sum(_gamma * _g, axis=1)
		G_mu /= self.n_components * F.sqrt(_w2)

		G_sig = F.sum(_gamma * (_g**2 - 1), axis=1)
		G_sig /= self.n_components * F.sqrt(2 * _w2)

		return F.concat([G_mu.reshape(n, -1), G_sig.reshape(n, -1)])

# class FVFunction(function.Function):
# 	def forward(self, inputs):
# 		import pdb; pdb.set_trace()

#	def backward(self, inputs):
#		import pdb; pdb.set_trace()


if __name__ == '__main__':

	from chainer.computational_graph import build_computational_graph
	class Model(chainer.Chain):
		def __init__(self):
			super(Model, self).__init__()
			with self.init_scope():
				self.link=FVLayer(2048, 5)

		def __call__(self, x):
			return self.link(x)

	model = Model()
	model.cleargrads()

	X = model.xp.ones((12, 3, 2048), dtype=np.float32)
	X = chainer.Variable(X)
	y = model(X)

	import pdb; pdb.set_trace()
	F.sum(y).backward()
	import pdb; pdb.set_trace()

	g = build_computational_graph([y])

	with open("graph.dot", "w") as out:
		out.write(g.dump())

	print(y)
